# -*- coding: utf-8 -*-
# Copyright (C) Alexander Pace, Tanner Prestegard,
#               Branson Stephens, Brian Moe (2020)
#
# This file is part of gracedb
#
# gracedb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gracedb.  If not, see <http://www.gnu.org/licenses/>

# Sources:
#  1) https://stackoverflow.com/questions/45539422/can-we-reload-a-page-url-
#     in-python-using-urllib-or-urllib2-or-requests-or-mechan

#  2) https://2.python-requests.org/en/master/user/advanced/#example-
#     specific-ssl-version

#  3) https://urllib3.readthedocs.io/en/1.2.1/pools.html

import logging

from functools import partial
from igwn_auth_utils import scitoken_authorization_header
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.connection import HTTPSConnection
from requests.packages.urllib3.connectionpool \
    import HTTPSConnectionPool, HTTPConnectionPool
from .cert import check_certificate_expiration
from .token import check_token_expiration, find_token_or_none

# Some variables:
auth_header = 'Authorization'

# initiate logging:
log = logging.getLogger(__name__)


class GraceDbCredAdapter(HTTPAdapter):
    def __init__(self, cert=None,
                 token=None, reload_buffer=0,
                 fail_if_noauth=False, **kwargs):
        super(GraceDbCredAdapter, self).__init__(**kwargs)
        https_pool_cls = partial(
            GraceDbCredHTTPSConnectionPool,
            cert=cert,
            token=token,
            reload_buffer=reload_buffer,
            fail_if_noauth=fail_if_noauth)

        self.poolmanager.pool_classes_by_scheme = {
            'http': HTTPConnectionPool,
            'https': https_pool_cls
        }


class GraceDbCredHTTPSConnection(HTTPSConnection):
    def __init__(self, host, cert=None,
                 token=None, reload_buffer=0, **kwargs):
        # At this point, te HTTPSConnection is initialized
        # but unconnected. Set this property to 'True'
        self.unestablished_connection = True
        super(GraceDbCredHTTPSConnection, self).__init__(host, **kwargs)

    @property
    def unestablished_connection(self):
        return self._unestablished_connection

    @unestablished_connection.setter
    def unestablished_connection(self, value):
        self._unestablished_connection = value

    def connect(self):
        # Connected. After this step, the unestablished
        # property is false.
        self.unestablished_connection = False
        super(GraceDbCredHTTPSConnection, self).connect()


class GraceDbCredHTTPSConnectionPool(HTTPSConnectionPool):
    # ConnectionPool object gets used in the HTTPAdapter.
    # "ConnectionCls" is a HTTP(S)COnnection object to use
    # As the underlying connection.

    # Source: https://urllib3.readthedocs.io/en/latest/
    #         reference/#module-urllib3.connectionpool

    ConnectionCls = GraceDbCredHTTPSConnection

    def __init__(self, host, port=None, cert=None,
                 token=None, reload_buffer=0,
                 fail_if_noauth=False, **kwargs):

        super(GraceDbCredHTTPSConnectionPool, self).__init__(
            host, port=port, **kwargs)

        self._cert = cert
        # self._token is a proxy for determining if this is a token auth
        # session and self._cur_token is the current token to validate
        # and refresh
        self._token = bool(token)
        self._cur_token = token
        self._reload_buffer = reload_buffer
        self._fail_if_noauth = fail_if_noauth

    def _expired_cred(self):
        if self._token:
            return check_token_expiration(
                self._cur_token,
                self._reload_buffer)
        elif self._cert:
            return check_certificate_expiration(
                self._cert,
                self._reload_buffer)
        else:
            raise ValueError('No renewable credentials (token or '
                             'certificate) provided.')

    def _get_conn(self, timeout=None):
        while True:
            # Start the connection object. At this step, the connection
            # unestablished variable is true
            conn = super(GraceDbCredHTTPSConnectionPool, self)._get_conn(
                timeout)

            # 'returning' the connection object then triggers the
            # connection to be established. Establish a new connection
            # if it's unestablished, or if the cert expiration is within the
            # reload buffer. Establishing the new connection will (hopefully
            # load the new cert.
            if conn.unestablished_connection or not self._expired_cred():
                if self._token:

                    # Attempt to get a new token. If no valid token is found,
                    # then None, and the adaptor will check for a valid token
                    # again on each subsequent request. if fail_if_noauth=true,
                    # then raise an error.
                    self._cur_token = find_token_or_none(
                        self.host,
                        fail_if_noauth=self._fail_if_noauth)

                    # Now update the headers. If there is no current token,
                    # pop the auth header to reflect the lack of a valid token:
                    if self._cur_token is None:
                        self.headers.pop(auth_header, None)
                    # Otherwise, update the auth header with the new token:
                    else:
                        self.headers.update({
                            auth_header:
                                scitoken_authorization_header(self._cur_token)
                        })

                return conn

            # otherwise, kill the connection which will reset unestablished_..
            # to true and then exit the loop.
            conn.close()



# The below code was heavily influenced from:
# https://code.activestate.com/recipes/577659-decorators-for-adding-aliases-to-methods-in-a-clas/
# , and was modified for ligo-gracedb.

# Please see the referenced page for authorship and copyright information.

import warnings
import functools


class alias(object):
    """
    Alias class that can be used as a decorator for making methods callable
    through other names (or "aliases").
    """

    def __init__(self, *aliases):
        self.aliases = set(aliases)

    def __call__(self, f):
        """
        Method call wrapper.
        """
        f._aliases = self.aliases
        return f


def aliased(aliased_class):
    """
    Decorator function that must be used in combination with @alias
    decorator. Modified to produce a deprecation warning.
    """
    def warning_wrapper(func, alias):

        warning_string = ("Method {old_name} has been replaced by {new_name}"
                          ", and will be deprecated in a future release")

        @functools.wraps(func)
        def wrapper(*args, **kwds):
            warnings.warn(warning_string.format(old_name=alias,
                                                new_name=func.__name__),
                          DeprecationWarning,
                          stacklevel=2)
            return func(*args, **kwds)
        return wrapper

    original_methods = aliased_class.__dict__.copy()
    for name, method in original_methods.items():
        if hasattr(method, '_aliases'):
            for alias in method._aliases - set(original_methods):
                wrapped_method = warning_wrapper(method, alias)
                setattr(aliased_class, alias, wrapped_method)
    return aliased_class



import os
import pytest
import tempfile
from pathlib import Path
from unittest import mock

from ligo.gracedb.rest import GraceDb
from igwn_auth_utils.error import IgwnAuthError


def test_provide_x509_cert_and_key():
    """Test client instantiation with provided certificate and key files"""
    # Set up cert and key files
    cert_file = '/tmp/cert_file'
    key_file = '/tmp/key_file'

    load_cert_func = 'ligo.gracedb.cert.load_certificate'
    with mock.patch(load_cert_func):  # noqa: E127
        # Initialize client
        g = GraceDb(cred=(cert_file, key_file))

    # Output credentials:
    creds = g.show_credentials(print_output=False)

    # Check credentials
    assert len(g.cert) == 2
    assert 'x509' in g.auth_type
    assert creds['x509'].get('cert_file') == cert_file
    assert creds['x509'].get('key_file') == key_file


def test_provide_x509_proxy():
    """Test client instantiation with provided combined proxy file"""
    # Set up combined proxy file
    proxy_file = '/tmp/proxy_file'

    exists_func = 'os.path.exists'
    with mock.patch(exists_func) as mock_exists_func:  # noqa: E127
        g = GraceDb(cred=proxy_file)

        mock_exists_func.return_value = True
        # Initialize client

    # Output credentials:
    creds = g.show_credentials(print_output=False)

    # Check credentials
    assert isinstance(g.cert, str)
    assert 'x509' in g.auth_type
    assert creds['x509'].get('cert_file') == proxy_file
    assert creds['x509'].get('key_file') == proxy_file


def test_provide_all_creds():
    """Test providing all credentials to the constructor
       FIXME: add scitokens to this"""
    # Setup
    cert_file = '/tmp/cert_file'
    key_file = '/tmp/key_file'

    # Initialize client
    g = GraceDb(cred=(cert_file, key_file))

    # Check credentials - should prioritze x509 credentials
    creds = g.show_credentials(print_output=False)

    assert len(g.cert) == 2
    assert 'x509' in g.auth_type
    assert creds['x509'].get('cert_file') == cert_file
    assert creds['x509'].get('key_file') == key_file


def test_x509_credentials_lookup():
    """Test lookup of X509 credentials"""
    # Setup
    cert_file = '/tmp/cert_file'
    key_file = '/tmp/key_file'

    g = GraceDb(cred=(cert_file, key_file))

    # the 'auth_type' variable doesn't get set in this case, but it's
    # sort of fictitious when self.cert is being mocked.
    assert len(g.cert) == 2


@mock.patch(
    "igwn_auth_utils.x509.validate_certificate",
    lambda cert, timeleft: True,
)
@mock.patch.dict(os.environ)
def test_x509_lookup_cert_key_from_envvars():
    """Test lookup of X509 cert and key from environment variables"""
    # FIXME: use tmp_path fixture when pytest min_version is 3.9.0
    with tempfile.TemporaryDirectory() as tmpdir:
        tmppath = Path(tmpdir)

        # Setup
        cert_file = tmppath / "cert_file"
        key_file = tmppath / "key_file"
        cert_file.touch()  # file needs to exist
        key_file.touch()  # file needs to exist
        os.environ["X509_USER_CERT"] = str(cert_file)
        os.environ["X509_USER_KEY"] = str(key_file)

        # Initialize client
        g = GraceDb()

    # Check credentials - should prioritze x509 credentials
    creds = g.show_credentials(print_output=False)

    assert len(g.cert) == 2
    assert 'x509' in g.auth_type
    assert creds['x509'].get('cert_file') == str(cert_file)
    assert creds['x509'].get('key_file') == str(key_file)


@mock.patch(
    "igwn_auth_utils.x509.validate_certificate",
    lambda cert, timeleft: True,
)
@mock.patch.dict(os.environ)
def test_x509_lookup_proxy_from_envvars():
    """Test lookup of X509 combined provxy file from environment variables"""
    # FIXME: use tmp_path fixture when pytest min_version is 3.9.0
    with tempfile.TemporaryDirectory() as tmpdir:
        # Setup
        proxy_file = Path(tmpdir) / "proxy_file"
        proxy_file.touch()
        os.environ.pop("X509_USER_CERT", None)
        os.environ.pop("X509_USER_KEY", None)
        os.environ["X509_USER_PROXY"] = str(proxy_file)

        # Initialize client
        g = GraceDb()

    # Check credentials - should prioritze x509 credentials
    creds = g.show_credentials(print_output=False)

    assert isinstance(g.cert, str)
    assert 'x509' in g.auth_type
    assert creds['x509'].get('cert_file') == str(proxy_file)
    assert creds['x509'].get('key_file') == str(proxy_file)


@mock.patch("igwn_auth_utils.requests.find_scitoken", return_value=None)
@mock.patch("igwn_auth_utils.requests.find_x509_credentials",
            return_value=None)
def test_no_credentials(_find_x509, _find_token):
    """Test client instantiation with no credentials at all"""
    # Check credentials
    g = GraceDb(fail_if_noauth=False)
    assert not g.cert
    assert not g.headers.get("Authorization")


@mock.patch.dict("os.environ", {
    "X509_USER_CERT": "cert_file",
    "X509_USER_KEY": "key_file",
})
def test_force_noauth():
    """Test forcing no authentication, even with X509 certs available"""
    # Initialize client
    g = GraceDb(force_noauth=True)

    # Check credentials
    assert not g.auth_type


@mock.patch("igwn_auth_utils.requests.find_scitoken", return_value=None)
@mock.patch("igwn_auth_utils.requests.find_x509_credentials",
            return_value=None)
def test_fail_if_noauth(*patches):
    """Test failing if no authentication credentials are provided"""
    err_str = 'no valid authorisation credentials found'
    with pytest.raises(IgwnAuthError, match=err_str):
        GraceDb(fail_if_noauth=True)


def test_fail_if_noauth_creds():
    """Test fail_if_noauth doesn't error if credentials are provided"""
    cert_file = '/tmp/cert_file'
    key_file = '/tmp/key_file'

    # Initialize client:
    g = GraceDb(cred=(cert_file, key_file), fail_if_noauth=True)

    # Check credentials
    assert g.cert == (cert_file, key_file)


def test_force_noauth_and_fail_if_noauth():
    # Initialize client
    err_str = 'cannot select both force_noauth and fail_if_noauth'
    with pytest.raises(ValueError, match=err_str):
        GraceDb(force_noauth=True, fail_if_noauth=True)


@pytest.mark.parametrize(
    "resource,key",
    [
        ('api_versions', 'api-versions'),
        ('server_version', 'server-version'),
        ('links', 'links'),
        ('templates', 'templates'),
        ('groups', 'groups'),
        ('pipelines', 'pipelines'),
        ('searches', 'searches'),
        ('allowed_labels', 'labels'),
        ('superevent_categories', 'superevent-categories'),
        ('em_groups', 'em-groups'),
        ('voevent_types', 'voevent-types'),
        ('signoff_types', 'signoff-types'),
        ('signoff_statuses', 'signoff-statuses'),
        ('instruments', 'instruments'),
    ]
)
def test_properties_from_api_root(safe_client, resource, key):
    si_prop = 'ligo.gracedb.rest.GraceDb.service_info'
    with mock.patch(si_prop, new_callable=mock.PropertyMock()) as mock_si:
        getattr(safe_client, resource)

    call_args, call_kwargs = mock_si.get.call_args
    assert mock_si.get.call_count == 1
    assert len(call_args) == 1
    assert call_kwargs == {}
    assert call_args[0] == key


@pytest.mark.parametrize("api_version", [1, 1.2, [], (), {}])
def test_bad_api_version(api_version):
    err_msg = 'api_version should be a string'
    with pytest.raises(TypeError, match=err_msg):
        GraceDb(api_version=api_version)


@pytest.mark.parametrize(
    "service_url,api_version",
    [
        ('test', None),
        ('test/', None),
        ('test', 'v1'),
        ('test/', 'v2'),
        ('test/', 'default'),
    ],
)
def test_set_service_url(safe_client, service_url, api_version):
    safe_client._set_service_url(service_url, api_version)

    # Construct expected service urls
    expected_service_url = service_url.rstrip('/') + '/'
    expected_versioned_service_url = expected_service_url

    if api_version and api_version != 'default':
        expected_versioned_service_url += api_version + '/'

    assert safe_client._service_url == expected_service_url
    assert safe_client._versioned_service_url == expected_versioned_service_url


def test_invalid_retries(safe_client):
    retries = -1
    expected_response = 'Invalid value of retries'
    with pytest.raises(ValueError, match=expected_response):
        GraceDb(retries=retries)


def test_invalid_retry_codes(safe_client):
    retry_codes = 503
    expected_response = 'retry_codes must be a list'
    with pytest.raises(ValueError, match=expected_response):
        GraceDb(retry_codes=retry_codes)



import os
import pytest
import scitokens
import tempfile
from unittest import mock

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key
from datetime import datetime, timezone
from igwn_auth_utils.scitokens import IgwnAuthError

from ligo.gracedb.rest import GraceDb
from ligo.gracedb.token import check_token_expiration, find_token_or_none

TEST_ISSUER = "local"
TEST_AUDIENCE = "ANY"
TEST_SCOPE = "gracedb.read"


# -- fixtures ---------------

@pytest.fixture(scope="session")  # one per suite is fine
def private_key():
    return generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend(),
    )


@pytest.fixture
def rtoken(private_key):
    """Create a token
    """
    # configure keycache
    from scitokens.utils import keycache
    kc = keycache.KeyCache.getinstance()
    kc.addkeyinfo(
        TEST_ISSUER,
        "test_key",
        private_key.public_key(),
        cache_timer=60,
    )

    # create token
    token = scitokens.SciToken(key=private_key, key_id="test_key")
    token.update_claims({
        "iss": TEST_ISSUER,
        "aud": TEST_AUDIENCE,
        "scope": TEST_SCOPE,
        "sub": 'albert.einstein@ligo.org',
    })
    serialized_token = token.serialize(issuer=TEST_ISSUER,
                                       lifetime=1000).decode("utf-8")
    return serialized_token


@pytest.fixture
def scitoken(private_key):
    """Create a token object
    """
    def _make_token(exp):
        # configure keycache
        from scitokens.utils import keycache
        kc = keycache.KeyCache.getinstance()
        kc.addkeyinfo(
            TEST_ISSUER,
            "test_key",
            private_key.public_key(),
            cache_timer=60,
        )

        # create token
        token = scitokens.SciToken(key=private_key, key_id="test_key")
        token.update_claims({
            "iss": TEST_ISSUER,
            "aud": TEST_AUDIENCE,
            "scope": TEST_SCOPE,
            "exp": exp,
            "sub": 'albert.einstein@ligo.org',
        })

        return token

    return _make_token


# -- test scitokens ---------

def test_scitokens(rtoken):
    with tempfile.TemporaryDirectory() as tmpdir:
        scitoken_file = os.path.join(tmpdir, "test_scitoken")

        with os.fdopen(os.open(scitoken_file,
                       os.O_RDWR | os.O_CREAT, 0o500), 'w+') as h:
            h.write(rtoken)

        # Initialize client
        with mock.patch.dict(os.environ, {'SCITOKEN_FILE': scitoken_file}):
            g = GraceDb()

        assert 'scitoken' in g.auth_type


def test_token_expiration(scitoken):
    # This strikes me as being kind of race-conditiony, but assign a
    # big buffer to make sure the token is within or outside
    # the validity period.
    valid_window = 1000
    now = int(datetime.now(timezone.utc).timestamp())

    # A valid token:
    assert check_token_expiration(scitoken(now + valid_window)) is False
    # An expired token:
    assert check_token_expiration(scitoken(now - valid_window)) is True
    # A valid token, but within the reload_buffer:
    assert check_token_expiration(scitoken(now - valid_window),
                                  reload_buffer=3600) is True


def test_finding_tokens(rtoken):
    host = "https://fakegracedb.ligo.org/"

    with tempfile.TemporaryDirectory() as tmpdir:
        scitoken_file = os.path.join(tmpdir, "test_scitoken")

        with os.fdopen(os.open(scitoken_file,
                       os.O_RDWR | os.O_CREAT, 0o500), 'w+') as h:
            h.write(rtoken)

        # Try finding the token:
        with mock.patch.dict(os.environ, {'SCITOKEN_FILE': scitoken_file}):
            assert isinstance(find_token_or_none(host),
                              scitokens.scitokens.SciToken)

        # temporarily remove any user's token if it exists, and then check
        # again:

        user_token_file = f'/tmp/bt_u{os.getuid()}'
        temp_token_file = '/tmp/.tmp_token'
        moved_file = os.path.exists(user_token_file)

        if moved_file:
            os.rename(user_token_file, temp_token_file)

        # Now check that the token can't be found:
        assert find_token_or_none(host) is None

        # Try again except raise an error:
        with pytest.raises(IgwnAuthError) as e:
            find_token_or_none(host, fail_if_noauth=True)

        assert "could not find a valid SciToken" in str(e.value)

        # Move the file back:
        if moved_file:
            os.rename(temp_token_file, user_token_file)


# This test is a WIP and will be improved in a future O4c release.
@pytest.mark.skip(reason="tested manually online, figure out this test")
def test_scitoken_reloading(rtoken):

    # Set up functions for tracking and mocking:
    set_up_conn_func = \
        'ligo.gracedb.adapter.GraceDbCredHTTPSConnection.connect'
    get_conn_func = \
        'requests.packages.urllib3.connectionpool.HTTPSConnectionPool'
    load_token_func = 'ligo.gracedb.token.find_token_or_none'
    token_expire_func = \
        'ligo.gracedb.token.check_token_expiration'

    # Set up the fake scitoken:
    with tempfile.TemporaryDirectory() as tmpdir:
        scitoken_file = os.path.join(tmpdir, "test_scitoken")

        with os.fdopen(os.open(scitoken_file,
                       os.O_RDWR | os.O_CREAT, 0o500), 'w+') as h:
            h.write(rtoken)

        with mock.patch(get_conn_func) as mock_get_conn, \
             mock.patch(set_up_conn_func) as mock_set_up_conn, \
             mock.patch(load_token_func) as mock_load_token, \
             mock.patch(token_expire_func) as mock_token_expire, \
             mock.patch.dict(os.environ, {'SCITOKEN_FILE': scitoken_file}):  # noqa: F841, E501

            # Now set up the client:
            g = GraceDb(reload_cred=True, reload_buffer=3600)

            # Confirm that it found the scitoken:
            assert 'scitoken' in g.auth_type

            # Make a call. Allow it to get past the typeerror to avoid
            # mocking the status-->response-->status_code
            g.get("https://fakeurl.com")

            # What should happen after this is, we should check that the
            # find_token_or_none, _expired_cred, check_token_expiration
            # functions were entered the expected number of times
            # right before the final request is mocked.



# -*- coding: utf-8 -*-
# Copyright (C) Alexander Pace, Duncan Meacher,
#               Duncan Macleod (2025)
#
# This file is part of gracedb
#
# gracedb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gracedb.  If not, see <http://www.gnu.org/licenses/>

# This file contains some helper tools and methods that aid with
# token renewal.

import datetime
import logging

from igwn_auth_utils.error import IgwnAuthError
from igwn_auth_utils.scitokens import find_token

log = logging.getLogger(__name__)

expired_message = ('token expiration ({exp}) -  current time ({utc})'
                   'is with the than reload_buffer ({buf}). reloading token.')


def check_token_expiration(token, reload_buffer=0):
    """ Checks to see if a token is expiring within the window
        defined by reload_buffer. If no token is found at all, then
        return false, prompting the client to reload. """

    log.debug('Checking scitoken expiration.')

    if token is None:
        log.info('A valid token was not found. Reloading token '
                 'adaptor.')
        return True
    else:
        exp_time = datetime.datetime.utcfromtimestamp(token['exp'])
        now = datetime.datetime.utcnow()
        expired = (exp_time - now) <= \
            datetime.timedelta(seconds=reload_buffer)
        if expired:
            log.info(expired_message.format(exp=exp_time,
                                            utc=now,
                                            buf=reload_buffer))

        return expired


def find_token_or_none(host, fail_if_noauth=False, scopes=None):
    """ Attempts to find a valid scitoken for the given host
        and scopes. If not, return None unless fail_if_noauth
        is true, then raise an error. """

    # Import the default token scope in the function to avoid
    # circular imports
    from .client import DEFAULT_TOKEN_SCOPE
    logging.debug('Finding a valid scitoken for the new connection.')

    # If no scopes were provided, then use the default:
    scopes = DEFAULT_TOKEN_SCOPE if scopes is None else scopes

    # Find a valid token for the current audience (host) and scope.
    # If no token can be found and fail_if_noauth is True, then
    # raise an error.
    try:
        return find_token(host, scope=scopes)
    except IgwnAuthError as err:
        if not fail_if_noauth:
            return None
        else:
            raise err



import json
import pytest
from unittest import mock


def test_gwtc_create_incompat_server(safe_client, gwtc_smap_content):
    number = '4'

    # Set up links mock
    mock_link = mock.MagicMock()
    mock_link_dict = {'not-gwtc': mock_link}
    link_prop = 'ligo.gracedb.rest.GraceDb.links'

    with mock.patch('ligo.gracedb.rest.GraceDb.post'), \
         mock.patch(link_prop, mock_link_dict):  # noqa: F841
        with pytest.raises(NotImplementedError):
            safe_client.gwtc_create(gwtc_smap_content, number)


def test_gwtc_create_input_bad_smap(safe_client):
    # Mock service info dict
    si_property = 'ligo.gracedb.rest.GraceDb.service_info'

    smap = 4
    number = '4'

    # Call:
    with mock.patch(si_property):
        with pytest.raises(ValueError):
            safe_client.gwtc_create(smap, number)


def test_gwtc_create_input_bad_number(safe_client, gwtc_smap_content):
    # Mock service info dict
    si_property = 'ligo.gracedb.rest.GraceDb.service_info'

    smap = gwtc_smap_content
    number = 14

    # Call:
    with mock.patch(si_property):
        with pytest.raises(ValueError):
            safe_client.gwtc_create(smap, number)


def test_gwtc_create_input_with_smap_filename(safe_client, gwtc_smap_content):
    # Make up a filename:
    smap = 'smap.json'
    number = '4'
    smap_superevent = list(gwtc_smap_content)[0]

    # Set up templates mock
    si_property = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_template = mock.MagicMock()
    mock_template_dict = {'gwtc': mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    # Set up mock open
    open_func = 'ligo.gracedb.rest.open'
    mock_data = json.dumps(gwtc_smap_content)
    open_mocker = mock.mock_open(read_data=mock_data)

    with mock.patch('ligo.gracedb.rest.GraceDb.post') as mock_post, \
         mock.patch(open_func, open_mocker), \
         mock.patch(si_property), \
         mock.patch(template_prop, mock_template_dict): # noqa E127

        safe_client.gwtc_create(smap, number)

        post_call_args, post_call_kwargs = mock_post.call_args
        assert 'json' in post_call_kwargs
        assert smap_superevent in post_call_kwargs['json']['smap']


def test_gwtc_create_input_with_smap_json(safe_client, gwtc_smap_content):
    number = '4'
    smap_superevent = list(gwtc_smap_content)[0]

    # Set up templates mock
    si_property = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_template = mock.MagicMock()
    mock_template_dict = {'gwtc': mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    with mock.patch('ligo.gracedb.rest.GraceDb.post') as mock_post, \
         mock.patch(si_property), \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127

        safe_client.gwtc_create(gwtc_smap_content, number)

        post_call_args, post_call_kwargs = mock_post.call_args
        assert 'json' in post_call_kwargs
        assert smap_superevent in post_call_kwargs['json']['smap']


def test_gwtc_get_incompat_server(safe_client):
    number = '4a'
    version = 1

    # Set up templates mock
    si_property = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_template = mock.MagicMock()
    mock_template_dict = {"not-gwtc-version-detail": mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    with mock.patch('ligo.gracedb.rest.GraceDb.get'), \
         mock.patch(si_property), \
         mock.patch(template_prop, mock_template_dict):  # noqa: F841
        with pytest.raises(NotImplementedError):
            safe_client.gwtc_get(number, version)


@pytest.mark.parametrize(
    "number,version",
    [
        (None, None),
        (4, None),
        ('4a', None),
        (None, '4'),
        ('4a', 'most_recent'),
        (4, 'latest'),
    ])
def test_gwtc_get_invalid_values(safe_client, number, version):
    # Set up templates mock
    si_property = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_template = mock.MagicMock()
    mock_template_dict = {"gwtc-version-detail": mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    with mock.patch('ligo.gracedb.rest.GraceDb.get'), \
         mock.patch(si_property), \
         mock.patch(template_prop, mock_template_dict):  # noqa: F841
        with pytest.raises(ValueError):
            safe_client.gwtc_get(number, version)


def test_gwtc_get(safe_client):
    number = '4a'
    version = 1

    # Set up templates mock
    si_property = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_template = mock.MagicMock()
    mock_template_dict = {"gwtc-version-detail": mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    with mock.patch('ligo.gracedb.rest.GraceDb.get') as mock_get, \
         mock.patch(si_property), \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.gwtc_get(number, version)

        get_call_args, get_call_kwargs = mock_get.call_args
        assert len(get_call_args) == 1
        assert get_call_kwargs == {}

# FIXME: mocking the generator object is non-zero effort and should
# be implemented in a future client release for gwtc_list


@pytest.mark.parametrize(
    "link_key", ['gwtc', 'not-gwtc']
)
def test_gwtc_numbers(safe_client, link_key):
    # Set up links mock
    si_property = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_link = mock.MagicMock()
    mock_link_dict = {link_key: mock_link}
    link_prop = 'ligo.gracedb.rest.GraceDb.links'

    with mock.patch('ligo.gracedb.rest.GraceDb.get') as mock_get, \
         mock.patch(si_property), \
         mock.patch(link_prop, mock_link_dict):  # noqa: F841

        if (link_key == 'gwtc'):
            safe_client.gwtc_numbers()
            get_call_args, get_call_kwargs = mock_get.call_args
            # no args to this method, but the get() passes the uri
            assert len(get_call_args) == 1
        else:
            with pytest.raises(NotImplementedError):
                safe_client.gwtc_numbers()


@pytest.mark.parametrize(
    "template_key,number", [
        ('gwtc-number-list', '4'),
        ('not-gwtc-number-list', '4'),
        ('gwtc-number-list', 4),
    ]
)
def test_gwtc_versions(safe_client, template_key, number):
    # Set up templates mock
    si_property = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_template = mock.MagicMock()
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    with mock.patch('ligo.gracedb.rest.GraceDb.get') as mock_get, \
         mock.patch(si_property), \
         mock.patch(template_prop, mock_template_dict):  # noqa: F841

        if (template_key == 'gwtc-number-list'):
            if isinstance(number, str):
                safe_client.gwtc_versions(number)
                get_call_args, get_call_kwargs = mock_get.call_args
                assert len(get_call_args) == 1
            else:
                with pytest.raises(ValueError):
                    safe_client.gwtc_versions(number)
        else:
            with pytest.raises(NotImplementedError):
                safe_client.gwtc_versions(number)



# this is necessary to maintain compatibility with legacy builders, in this
# case, the py3 build macros in redhat packaging.

# https://setuptools.pypa.io/en/latest/userguide/pyproject_config.html

from setuptools import setup

setup()



# -*- coding: utf-8 -*-
# Copyright (C) Alexander Pace, Tanner Prestegard,
#               Branson Stephens, Brian Moe (2020)
#
# This file is part of gracedb
#
# gracedb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gracedb.  If not, see <http://www.gnu.org/licenses/>.
import json as json_lib
import warnings
from datetime import datetime
from pprint import pprint
from requests.adapters import HTTPAdapter, Retry
from requests import codes
from urllib.parse import urlparse

from .version import __version__
from .adapter import GraceDbCredAdapter
from .utils import hook_response, raise_status_exception

from igwn_auth_utils.requests import (HTTPSciTokenAuth, Session)

DEFAULT_SERVICE_URL = "https://gracedb.ligo.org/api/"
DEFAULT_RETRY_CODES = [
    codes.BAD_GATEWAY,      # 502
    codes.GATEWAY_TIMEOUT,  # 504
    codes.REQUEST_TIMEOUT,  # 408
    codes.CONFLICT,         # 409
]

DEFAULT_ALLOWED_METHODS = None
DEFAULT_TOKEN_SCOPE = "gracedb.read"


class GraceDBClient(Session):
    """
    url (:obj:`str`, optional): URL of server API
    cred (:obj:`tuple` or :obj:`str, optional): a tuple or list of
        (``/path/to/cert/file``, ``/path/to/key/file) or a single path to
        a combined proxy file (if using an X.509 certificate for
        authentication)
    force_noauth (:obj:`bool`, optional): set to True if you want to skip
        credential lookup and use this client as an unauthenticated user
    fail_if_noauth (:obj:`bool`, optional): set to True if you want the
        constructor to fail if no authentication credentials are provided
        or found, or if no valid scitoken credentials were found when
        using the ``reload_cred`` option
    reload_certificate (:obj:`bool`, optional): if ``True``, your
        certificate will be checked before each request whether it is
        within ``reload_buffer`` seconds of expiration, and if so, it will
        be reloaded. Useful for processes which may live longer than the
        certificate lifetime and have an automated method for certificate
        renewal. The path to the new/renewed certificate **must** be the
        same as for the old certificate. To be deprecated in favor of
        ``reload_cred``.
    reload_cred (:obj:`bool`, optional): if ``True``, the user's token
        or certificate will be checked before each request if it is within
        ``reload_buffer`` seconds of expiration, and if so, will be
        reloaded. Used for long-lived processes in conjunction with a
        user's own credential reloading mechanism.
    reload_buffer (:obj:`int`, optional): buffer (in seconds) for reloading
        a certificate in advance of its expiration. Only used if
        ``reload_certificate`` is ``True``.
    retries (:obj:`int`, optional): the maximum number of retries to
        attempt on an error from the server. Default 5.
    backoff_factor (:obj:`float`, optional): The backoff factor for retrying.
        Default: 0.1. Refer to urllib3 documentation for usage.
    retry_codes (:obj:`list`, optional): List of HTTPError codes to retry
        on. Default: [502, 504, 408, 409].
    allowed_methods (:obj:`list` or None, optional): whitelist of allowed
        http verbs to retry. Default: None (all verbs).
    use_auth: (:obj:`str`, optional): Specify the authentication method
        to be used. Choices: ``all``, ``scitoken``, ``x509``. Default:
        ``all``.

    Authentication details:

    By default, the code will search for available scitokens, and then X.509
    certificates. The credentials will then be used in that order of
    preference unless overridden by the ``use_auth`` variable.

    To use an X.509 certificate:
        Provide a path to an X.509 certificate and key or a single
           combined proxy file using the ``cred`` parameter.
    Or:
        The code will look for a certificate in a default location
            (``/tmp/x509up_u%d``, where ``%d`` is your user ID)
    """

    def __init__(self, url=DEFAULT_SERVICE_URL, cred=None,
                 reload_certificate=False,
                 reload_cred=False, reload_buffer=300,
                 use_auth='all', retries=5,
                 retry_codes=DEFAULT_RETRY_CODES,
                 allowed_methods=DEFAULT_ALLOWED_METHODS,
                 backoff_factor=0.1, *args, **kwargs):

        # Set which auth method to use if specified, otherwise use all.
        if use_auth == 'x509':
            if cred is None:
                cred = True
            kwargs.setdefault("token", False)
        elif use_auth == 'scitoken':
            kwargs.setdefault("token", True)
            if cred is None:
                cred = False
        kwargs.setdefault("token_scope", DEFAULT_TOKEN_SCOPE)

        super().__init__(cert=cred, url=url, *args, **kwargs)

        # Initialize variables:
        self.host = urlparse(url).hostname
        self._fail_if_noauth = kwargs.get('fail_if_noauth')

        # Define auth_type based on.... type of auth.
        self.auth_type = set()
        if use_auth != "x509" and isinstance(self.auth, HTTPSciTokenAuth):
            self.auth_type.add("scitoken")
        if use_auth != 'scitoken' and self.cert:
            self.auth_type.add("x509")

        # Update session headers:
        self.headers.update(self._update_headers())

        # Adjust the response via a session hook:
        self.hooks = {'response': [hook_response, raise_status_exception]}

        # Add the retrying adaptor:
        # https://stackoverflow.com/a/35504626
        # Sanity check the inputs:

        # 'retries' must be a positive (or zero) integer:
        if (not isinstance(retries, int) or retries < 0):
            raise ValueError('Invalid value of retries')

        # 'retry_codes' must be a list:
        if not isinstance(retry_codes, list):
            raise ValueError('retry_codes must be a list')

        if retries > 0:
            # the allowed_methods option was added as of urllib3==1.26.1
            # and replaces method_whitelist:
            if hasattr(Retry.DEFAULT, 'allowed_methods'):
                retries = Retry(total=retries,
                                backoff_factor=backoff_factor,
                                status_forcelist=retry_codes,
                                allowed_methods=allowed_methods)
            else:
                retries = Retry(total=retries,
                                backoff_factor=backoff_factor,
                                status_forcelist=retry_codes,
                                method_whitelist=allowed_methods)

        if reload_cred and isinstance(self.auth, HTTPSciTokenAuth) and \
                use_auth != 'x509':
            # Get the token object to pass to the reloading adapter.
            # Fail with an error if a token cannot be found when
            # fail_if_noauth is "true"
            token = self.auth.find_token(self.host, error=self._fail_if_noauth)
            self.mount('https://', GraceDbCredAdapter(
                       token=token,
                       fail_if_noauth=self._fail_if_noauth,
                       reload_buffer=reload_buffer,
                       max_retries=retries))
        elif (reload_certificate or reload_cred) and self.cert:
            if reload_certificate:
                warning_text = ('reload_certificate has been deprecated in '
                                'favor of reload_cred and will be removed in '
                                'a future release.')
                warnings.warn(warning_text,
                              DeprecationWarning,
                              stacklevel=2)
            self.mount('https://', GraceDbCredAdapter(
                       cert=self.cert,
                       reload_buffer=reload_buffer,
                       max_retries=retries))
        else:
            self.mount('https://', HTTPAdapter(max_retries=retries))

    def _update_headers(self):
        """ Update the sessions' headers:
        """
        new_headers = {}
        # Assign the user agent. This shows up in gracedb's log files.
        new_headers.update({'User-Agent':
                            'gracedb-client/{}'.format(__version__)})
        new_headers.update({'Accept-Encoding':
                            'gzip, deflate'})
        return new_headers

    # hijack 'Session.request':
    # https://2.python-requests.org/en/master/api/#requests.Session.request
    def request(
            self, method, url, params=None, data=None, headers=None,
            cookies=None, files=None, auth=None, timeout=None,
            allow_redirects=True, proxies=None, hooks=None, stream=None,
            verify=None, cert=None, json=None):
        return super().request(
            method, url, params=params, data=data, headers=headers,
            cookies=cookies, files=files, auth=auth,
            timeout=timeout, allow_redirects=True, proxies=proxies,
            hooks=hooks, stream=stream, verify=verify, cert=cert, json=json)

    # Extra definitions to return closed contexts' connections
    # back to the pool:
    # https://stackoverflow.com/questions/48160728/resourcewarning
    # -unclosed-socket-in-python-3-unit-test
    def close(self):
        super().close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    # For getting files, return a "raw" file-type object.
    # Automatically decode content.
    def get_file(self, url, **kwargs):
        resp = self.get(url, stream=True, **kwargs)
        resp.raw.decode_content = True
        resp.raw.status_code = resp.status_code
        resp.raw.json = resp.json
        return resp.raw

    # Return client credentials
    def show_credentials(self, print_output=True):
        """ Prints authentication type and credentials info."""
        output = {}
        if not self.auth_type:
            output = {'auth_type(s)': 'No auth type found'}

        if 'scitoken' in self.auth_type:
            token = self.auth.find_token(self.host, error=False)
            if token is not None:
                exp = datetime.utcfromtimestamp(token['exp']).isoformat()
                output['scitoken'] = {
                    'SciToken subject': token['sub'],
                    'SciToken expiration': exp,
                    'SciToken scope': token['scope'],
                    'SciToken audience': token['aud'],
                }
        if 'x509' in self.auth_type:
            if isinstance(self.cert, tuple):
                cert = {'cert_file': self.cert[0],
                        'key_file': self.cert[1]}
            elif isinstance(self.cert, str):
                cert = {'cert_file': self.cert,
                        'key_file': self.cert}
            else:
                raise ValueError("Problem reading authentication certificate")
            output['x509'] = cert

        if print_output:
            pprint(output)
        else:
            return output

    def get_user_info(self):
        """Get information from the server about your user account."""
        user_info_link = self.links.get('user-info', None)
        if user_info_link is None:
            raise RuntimeError('Server does not provide a user info endpoint')
        return self.get(user_info_link)

    @classmethod
    def load_json_from_response(cls, response):
        """ Always return a json content response, even when the server
            provides a 204: no content"""
        # Check if response exists:
        if not response:
            raise ValueError("No response object provided")

        # Check if there is response content. If not, create it.
        if response.content == 'No Content':
            response_content = '{}'

        # Some responses send back strings of strings. This iterates
        # until proper dict is returned, or if it doesn't make progress.
        num_tries = 1
        response_content = response.content.decode('utf-8')

        while isinstance(response_content, str) and num_tries < 3:
            response_content = json_lib.loads(response_content)
            num_tries += 1

        if isinstance(response_content, dict):
            return response_content
        else:
            return ValueError("ERROR: got unexpected content from "
                              "the server: {}".format(response_content))
